import abc
from collections.abc import Iterator

import torch


class OptimizerFactory(abc.ABC):
    """
    Interface for optimizer factory
    """

    @abc.abstractmethod
    def create(self, parameters: Iterator[torch.nn.Parameter]) -> torch.optim.Optimizer:  # pragma: no cover
        """
        Creates optimizer based on parameters.

        :param parameters: torch parameters to initialize optimizer

        :returns: torch optimizer
        """


class LRSchedulerFactory(abc.ABC):
    """
    Interface for learning rate scheduler factory
    """

    @abc.abstractmethod
    def create(self, optimizer: torch.optim.Optimizer) -> torch.optim.lr_scheduler._LRScheduler:  # pragma: no cover
        """
        Creates learning rate scheduler based on optimizer.

        :param optimizer: torch optimizer

        :returns: torch LRScheduler
        """


class FatOptimizerFactory(OptimizerFactory):
    """
    Factory that creates optimizer depending on passed parameters
    """

    def __init__(
        self,
        optimizer: str = "adam",
        learning_rate: float = 0.001,
        weight_decay: float = 0.0,
        sgd_momentum: float = 0.0,
        betas: tuple[float, float] = (0.9, 0.98),
    ) -> None:
        super().__init__()
        self.optimizer = optimizer
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.sgd_momentum = sgd_momentum
        self.betas = betas

    def create(self, parameters: Iterator[torch.nn.Parameter]) -> torch.optim.Optimizer:
        """
        Creates optimizer based on parameters.

        :param parameters: torch parameters to initialize optimizer

        :returns: torch optimizer
        """
        if self.optimizer == "adam":
            return torch.optim.Adam(parameters, lr=self.learning_rate, weight_decay=self.weight_decay, betas=self.betas)
        if self.optimizer == "sgd":
            return torch.optim.SGD(
                parameters, lr=self.learning_rate, weight_decay=self.weight_decay, momentum=self.sgd_momentum
            )

        msg = "Unexpected optimizer"
        raise ValueError(msg)


class FatLRSchedulerFactory(LRSchedulerFactory):
    """
    Factory that creates learning rate schedule depending on passed parameters
    """

    def __init__(
        self,
        decay_step: int = 25,
        gamma: float = 1.0,
    ) -> None:
        super().__init__()
        self.decay_step = decay_step
        self.gamma = gamma

    def create(self, optimizer: torch.optim.Optimizer) -> torch.optim.lr_scheduler._LRScheduler:
        """
        Creates learning rate scheduler based on optimizer.

        :param optimizer: torch optimizer

        :returns: torch LRScheduler
        """
        return torch.optim.lr_scheduler.StepLR(optimizer, step_size=self.decay_step, gamma=self.gamma)



from collections import defaultdict
from typing import Union

import numpy as np
import polars as pl

from replay.utils import PYSPARK_AVAILABLE, PandasDataFrame, PolarsDataFrame, SparkDataFrame

from .base_metric import (
    Metric,
    MetricsDataFrameLike,
    MetricsMeanReturnType,
    MetricsPerUserReturnType,
    MetricsReturnType,
)
from .descriptors import CalculationDescriptor, Mean

if PYSPARK_AVAILABLE:
    from pyspark.sql import (
        Window,
        functions as sf,
    )


class CategoricalDiversity(Metric):
    """
    Metric calculation is as follows:

    * take ``K`` recommendations with the biggest ``score`` for each ``user_id``
    * count the number of distinct ``category_id`` in these recommendations / ``K``
    * average this number for all users

    >>> category_recommendations
       query_id  category_id  rating
    0         1            3    0.6
    1         1            7    0.5
    2         1           10    0.4
    3         1           11    0.3
    4         1            2    0.2
    5         2            5    0.6
    6         2            8    0.5
    7         2           11    0.4
    8         2            1    0.3
    9         2            3    0.2
    10        3            4    1.0
    11        3            9    0.5
    12        3            2    0.1
    >>> from replay.metrics import Median, ConfidenceInterval, PerUser
    >>> CategoricalDiversity([3, 5])(category_recommendations)
    {'CategoricalDiversity@3': 1.0, 'CategoricalDiversity@5': 0.8666666666666667}
    >>> CategoricalDiversity([3, 5], mode=PerUser())(category_recommendations)
    {'CategoricalDiversity-PerUser@3': {1: 1.0, 2: 1.0, 3: 1.0},
     'CategoricalDiversity-PerUser@5': {1: 1.0, 2: 1.0, 3: 0.6}}
    >>> CategoricalDiversity([3, 5], mode=Median())(category_recommendations)
    {'CategoricalDiversity-Median@3': 1.0,
     'CategoricalDiversity-Median@5': 1.0}
    >>> CategoricalDiversity([3, 5], mode=ConfidenceInterval(alpha=0.95))(category_recommendations)
    {'CategoricalDiversity-ConfidenceInterval@3': 0.0,
     'CategoricalDiversity-ConfidenceInterval@5': 0.2613285312720073}
    <BLANKLINE>
    """

    def __init__(
        self,
        topk: Union[list, int],
        query_column: str = "query_id",
        category_column: str = "category_id",
        rating_column: str = "rating",
        mode: CalculationDescriptor = Mean(),
    ):
        """
        :param topk: (list or int): Consider the highest k scores in the ranking.
        :param user_column: (str): The name of the user column.
        :param category_column: (str): The name of the category column.
        :param score_column: (str): The name of the score column.
        :param mode: (CalculationDescriptor): class for calculating aggregation metrics.
            Default: ``Mean``.
        """
        super().__init__(
            topk=topk,
            query_column=query_column,
            rating_column=rating_column,
            mode=mode,
        )
        self.category_column = category_column

    def __call__(self, recommendations: MetricsDataFrameLike) -> MetricsReturnType:
        """
        Compute metric.

        :param recommendations: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict):
            model predictions.
            If DataFrame then it must contains user, category and score columns.
            If dict then key represents user_ids, value represents list of tuple(category, score).

        :return: metric values
        """
        if isinstance(recommendations, SparkDataFrame):
            return self._spark_call(recommendations)
        if isinstance(recommendations, PolarsDataFrame):
            return self._polars_call(recommendations)
        is_pandas = isinstance(recommendations, PandasDataFrame)
        recommendations = (
            self._convert_pandas_to_dict_with_score(recommendations)
            if is_pandas
            else self._convert_dict_to_dict_with_score(recommendations)
        )
        precalculated_answer = self._precalculate_unique_cats(recommendations)
        return self._dict_call(precalculated_answer)

    def _get_enriched_recommendations(
        self,
        recommendations: Union[PolarsDataFrame, SparkDataFrame],
    ) -> Union[PolarsDataFrame, SparkDataFrame]:
        if isinstance(recommendations, SparkDataFrame):
            return self._get_enriched_recommendations_spark(recommendations)
        else:
            return self._get_enriched_recommendations_polars(recommendations)

    def _get_enriched_recommendations_spark(self, recommendations: SparkDataFrame) -> SparkDataFrame:
        window = Window.partitionBy(self.query_column).orderBy(sf.col(self.rating_column).desc())
        sorted_by_score_recommendations = recommendations.withColumn("rank", sf.row_number().over(window))
        return sorted_by_score_recommendations

    def _get_enriched_recommendations_polars(self, recommendations: PolarsDataFrame) -> PolarsDataFrame:
        sorted_by_score_recommendations = recommendations.select(
            pl.all().sort_by(self.rating_column, descending=True).over(self.query_column)
        )
        sorted_by_score_recommendations = sorted_by_score_recommendations.with_columns(
            sorted_by_score_recommendations.select(
                pl.col(self.query_column).cum_count().over(self.query_column).alias("rank")
            )
        )
        return sorted_by_score_recommendations

    def _spark_compute_per_user(self, recs: SparkDataFrame) -> MetricsPerUserReturnType:
        distribution_per_user = defaultdict(list)
        for k in self.topk:
            filtered_recs = recs.filter(sf.col("rank") <= k)
            aggreagated_by_user = filtered_recs.groupBy(self.query_column).agg(sf.countDistinct(self.category_column))
            aggreagated_by_user_dict = aggreagated_by_user.rdd.collectAsMap()
            for user, metric in aggreagated_by_user_dict.items():
                distribution_per_user[user].append(metric / k)
        return self._aggregate_results_per_user(dict(distribution_per_user))

    def _polars_compute_per_user(self, recs: PolarsDataFrame) -> MetricsPerUserReturnType:
        distribution_per_user = defaultdict(list)
        for k in self.topk:
            filtered_recs = recs.filter(pl.col("rank") <= k)
            aggreagated_by_user = filtered_recs.group_by(self.query_column).agg(pl.col(self.category_column).n_unique())
            aggreagated_by_user_dict = dict(aggreagated_by_user.iter_rows())
            for user, metric in aggreagated_by_user_dict.items():
                distribution_per_user[user].append(metric / k)
        return self._aggregate_results_per_user(dict(distribution_per_user))

    def _spark_compute_agg(self, recs: SparkDataFrame) -> MetricsMeanReturnType:
        metrics = []
        for k in self.topk:
            filtered_recs = recs.filter(sf.col("rank") <= k)
            aggregated_by_user = (
                filtered_recs.groupBy(self.query_column)
                .agg(sf.countDistinct(self.category_column))
                .drop(self.query_column)
            )
            metrics.append(self._mode.spark(aggregated_by_user) / k)
        return self._aggregate_results(metrics)

    def _polars_compute_agg(self, recs: PolarsDataFrame) -> MetricsMeanReturnType:
        metrics = []
        for k in self.topk:
            filtered_recs = recs.filter(pl.col("rank") <= k)
            aggregated_by_user = (
                filtered_recs.group_by(self.query_column)
                .agg(pl.col(self.category_column).n_unique())
                .drop(self.query_column)
            )
            metrics.append(self._mode.cpu(aggregated_by_user) / k)
        return self._aggregate_results(metrics)

    def _spark_call(self, recommendations: SparkDataFrame) -> MetricsReturnType:
        """
        Implementation for Pyspark DataFrame.
        """
        recs = self._get_enriched_recommendations(recommendations)
        if self._mode.__name__ == "PerUser":
            return self._spark_compute_per_user(recs)
        return self._spark_compute_agg(recs)

    def _polars_call(self, recommendations: PolarsDataFrame) -> MetricsReturnType:
        """
        Implementation for Polars DataFrame.
        """
        recs = self._get_enriched_recommendations(recommendations)
        if self._mode.__name__ == "PerUser":
            return self._polars_compute_per_user(recs)
        return self._polars_compute_agg(recs)

    def _convert_pandas_to_dict_with_score(self, data: PandasDataFrame) -> dict:
        return (
            data.sort_values(by=self.rating_column, ascending=False)
            .groupby(self.query_column)[self.category_column]
            .apply(list)
            .to_dict()
        )

    def _precalculate_unique_cats(self, recommendations: dict) -> dict:
        """
        Precalculate unique categories for each prefix for each user.
        """
        answer = {}
        for user, cats in recommendations.items():
            unique = set()
            unique_len = []
            for cat in cats:
                unique.add(cat)
                unique_len.append(len(unique))
            answer[user] = unique_len
        return answer

    def _dict_compute_per_user(self, precalculated_answer: dict) -> MetricsPerUserReturnType:
        distribution_per_user = defaultdict(list)
        for k in self.topk:
            for user, unique_cats in precalculated_answer.items():
                distribution_per_user[user].append(unique_cats[min(len(unique_cats), k) - 1] / k)
        return self._aggregate_results_per_user(distribution_per_user)

    def _dict_compute_mean(self, precalculated_answer: dict) -> MetricsMeanReturnType:
        distribution_list = []
        for unique_cats in precalculated_answer.values():
            metrics_per_user = []
            for k in self.topk:
                metric = unique_cats[min(len(unique_cats), k) - 1] / k
                metrics_per_user.append(metric)
            distribution_list.append(metrics_per_user)

        distribution = np.stack(distribution_list)
        assert distribution.shape[1] == len(self.topk)
        metrics = [self._mode.cpu(distribution[:, k]) for k in range(distribution.shape[1])]
        return self._aggregate_results(metrics)

    def _dict_call(self, precalculated_answer: dict) -> MetricsReturnType:
        """
        Calculating metrics in dict format.
        """
        if self._mode.__name__ == "PerUser":
            return self._dict_compute_per_user(precalculated_answer)
        return self._dict_compute_mean(precalculated_answer)

    @staticmethod
    def _get_metric_value_by_user(ks: list[int], *args: list) -> list[float]:  # pragma: no cover
        pass



from replay.utils import TORCH_AVAILABLE

if TORCH_AVAILABLE:
    from .dataset import (
        Bert4RecPredictionBatch,
        Bert4RecPredictionDataset,
        Bert4RecTrainingBatch,
        Bert4RecTrainingDataset,
        Bert4RecUniformMasker,
        Bert4RecValidationBatch,
        Bert4RecValidationDataset,
    )
    from .lightning import Bert4Rec
    from .model import Bert4RecModel



"""
This module contains tools for preprocessing data including:

- filters
- processors for feature transforms
"""

from .converter import CSRConverter
from .discretizer import (
    Discretizer,
    GreedyDiscretizingRule,
    HandleInvalidStrategies,
    QuantileDiscretizingRule,
)
from .history_based_fp import (
    ConditionalPopularityProcessor,
    EmptyFeatureProcessor,
    HistoryBasedFeaturesProcessor,
    LogStatFeaturesProcessor,
)
from .label_encoder import LabelEncoder, LabelEncoderPartialFitWarning, LabelEncodingRule, SequenceEncodingRule
from .sessionizer import Sessionizer



from .base_metric import Metric


class Recall(Metric):
    """
    Recall measures the coverage of the recommended items, and is defined as:

    Mean percentage of relevant items, that was shown among top ``K`` recommendations.

    .. math::
        Recall@K(i) = \\frac {\\sum_{j=1}^{K}\\mathbb{1}_{r_{ij}}}{|Rel_i|}

    .. math::
        Recall@K = \\frac {\\sum_{i=1}^{N}Recall@K(i)}{N}

    :math:`\\mathbb{1}_{r_{ij}}` -- indicator function showing that user :math:`i` interacted with item :math:`j`

    :math:`|Rel_i|` -- the number of relevant items for user :math:`i`

    >>> recommendations
       query_id  item_id  rating
    0         1        3    0.6
    1         1        7    0.5
    2         1       10    0.4
    3         1       11    0.3
    4         1        2    0.2
    5         2        5    0.6
    6         2        8    0.5
    7         2       11    0.4
    8         2        1    0.3
    9         2        3    0.2
    10        3        4    1.0
    11        3        9    0.5
    12        3        2    0.1
    >>> groundtruth
       query_id  item_id
    0         1        5
    1         1        6
    2         1        7
    3         1        8
    4         1        9
    5         1       10
    6         2        6
    7         2        7
    8         2        4
    9         2       10
    10        2       11
    11        3        1
    12        3        2
    13        3        3
    14        3        4
    15        3        5
    >>> from replay.metrics import Median, ConfidenceInterval, PerUser
    >>> Recall(2)(recommendations, groundtruth)
    {'Recall@2': 0.12222222222222223}
    >>> Recall(2, mode=PerUser())(recommendations, groundtruth)
    {'Recall-PerUser@2': {1: 0.16666666666666666, 2: 0.0, 3: 0.2}}
    >>> Recall(2, mode=Median())(recommendations, groundtruth)
    {'Recall-Median@2': 0.16666666666666666}
    >>> Recall(2, mode=ConfidenceInterval(alpha=0.95))(recommendations, groundtruth)
    {'Recall-ConfidenceInterval@2': 0.12125130695058273}
    <BLANKLINE>
    """

    @staticmethod
    def _get_metric_value_by_user(ks: list[int], ground_truth: list, pred: list) -> list[float]:
        if not ground_truth or not pred:
            return [0.0 for _ in ks]
        set_gt = set(ground_truth)
        res = []
        for k in ks:
            ans = len(set(pred[:k]) & set(ground_truth))
            res.append(ans / len(set_gt))
        return res



import math
from typing import Optional

from scipy.optimize import root_scalar

from replay.utils import PYSPARK_AVAILABLE

from .ucb import UCB

if PYSPARK_AVAILABLE:
    from pyspark.sql.functions import udf
    from pyspark.sql.types import DoubleType


class KLUCB(UCB):
    """
    Bernoulli `bandit model
    <https://en.wikipedia.org/wiki/Multi-armed_bandit>`_. Same to :class:`UCB`
    computes item relevance as an upper confidence bound of true fraction of
    positive interactions.

    In a nutshell, KL-UCB considers the data as the history of interactions
    with items. The interaction may be either positive or negative. For each
    item the model computes empirical frequency of positive interactions
    and estimates the true frequency with an upper confidence bound. The higher
    the bound for an item is the more relevant it is presumed.

    The upper bound below is what differs from the classical UCB. It is
    computed according to the `original article
    <https://arxiv.org/pdf/1102.2490.pdf>`_ where is proven to produce
    assymptotically better results.

    .. math::
        u_i = \\max q \\in [0,1] :
        n_i \\cdot \\operatorname{KL}\\left(\\frac{p_i}{n_i}, q \\right)
        \\leqslant \\log(n) + c \\log(\\log(n)),

    where

    :math:`u_i` -- upper bound for item :math:`i`,

    :math:`c` -- exploration coeficient,

    :math:`n` -- number of interactions in log,

    :math:`n_i` -- number of interactions with item :math:`i`,

    :math:`p_i` -- number of positive interactions with item :math:`i`,

    and

    .. math::
        \\operatorname{KL}(p, q)
        = p \\log\\frac{p}{q} + (1-p)\\log\\frac{1-p}{1-q}

    is the KL-divergence of Bernoulli distribution with parameter :math:`p`
    from Bernoulli distribution with parameter :math:`q`.

    Being a bit trickier though, the bound shares with UCB the same
    exploration-exploitation tradeoff dilemma. You may increase the `c`
    coefficient in order to shift the tradeoff towards exploration or decrease
    it to set the model to be more sceptical of items with small volume of
    collected statistics. The authors of the `article
    <https://arxiv.org/pdf/1102.2490.pdf>`_ though claim `c = 0` to be of the
    best choice in practice.


    As any other RePlay model, KL-UCB takes a log to fit on as a ``DataFrame``
    with columns ``[user_idx, item_idx, timestamp, relevance]``. Following the
    procedure above, KL-UCB would see each row as a record of an interaction
    with ``item_idx`` with positive (relevance = 1) or negative (relevance = 0)
    outcome. ``user_idx`` and ``timestamp`` are ignored i.e. the model treats
    log as non-personalized - item scores are same for all users.

    If ``relevance`` column is not of 0/1 initially, then you have to decide
    what kind of relevance has to be considered as positive and convert
    ``relevance`` to binary format during preprocessing.

    To provide a prediction, KL-UCB would sample a set of recommended items for
    each user with probabilites proportional to obtained relevances.

    >>> import pandas as pd
    >>> from replay.data.dataset import Dataset, FeatureSchema, FeatureInfo, FeatureHint, FeatureType
    >>> from replay.utils.spark_utils import convert2spark
    >>> data_frame = pd.DataFrame({"user_id": [1, 2, 3, 3], "item_id": [1, 2, 1, 2], "rating": [1, 0, 0, 0]})
    >>> interactions = convert2spark(data_frame)
    >>> feature_schema = FeatureSchema(
    ...     [
    ...         FeatureInfo(
    ...             column="user_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.QUERY_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="item_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.ITEM_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="rating",
    ...             feature_type=FeatureType.NUMERICAL,
    ...             feature_hint=FeatureHint.RATING,
    ...         ),
    ...     ]
    ... )
    >>> dataset = Dataset(feature_schema, interactions)
    >>> model = KLUCB()
    >>> model.fit(dataset)
    >>> model.predict(dataset, k=2, queries=[1,2,3,4], items=[1,2,3]
    ... ).toPandas().sort_values(["user_id","rating","item_id"],
    ... ascending=[True,False,True]).reset_index(drop=True)
        user_id   item_id     rating
    0         1         3   1.000000
    1         1	        2   0.750000
    2         2	        3   1.000000
    3         2	        1   0.933013
    4         3	        3   1.000000
    5         4	        3   1.000000
    6         4	        1   0.933013

    """

    def __init__(
        self,
        exploration_coef: float = 0.0,
        sample: bool = False,
        seed: Optional[int] = None,
    ):
        """
        :param exploration_coef: exploration coefficient
        :param sample: flag to choose recommendation strategy.
            If True, items are sampled with a probability proportional
            to the calculated predicted relevance.
            Could be changed after model training by setting the `sample`
            attribute.
        :param seed: random seed. Provides reproducibility if fixed
        """

        super().__init__(exploration_coef, sample, seed)

    def _calc_item_popularity(self):
        right_hand_side = math.log(self.full_count) + self.coef * math.log(math.log(self.full_count))
        eps = 1e-12

        def bernoulli_kl(proba_p, proba_q):  # pragma: no cover
            return proba_p * math.log(proba_p / proba_q) + (1 - proba_p) * math.log((1 - proba_p) / (1 - proba_q))

        @udf(returnType=DoubleType())
        def get_ucb(pos, total):  # pragma: no cover
            proba = pos / total

            if proba == 0:
                ucb = root_scalar(
                    f=lambda qq: math.log(1 / (1 - qq)) - right_hand_side, bracket=[0, 1 - eps], method="brentq"
                ).root
                return ucb

            if proba == 1:
                ucb = root_scalar(
                    f=lambda qq: math.log(1 / qq) - right_hand_side, bracket=[0 + eps, 1], method="brentq"
                ).root
                return ucb

            ucb = root_scalar(
                f=lambda q: total * bernoulli_kl(proba, q) - right_hand_side, bracket=[proba, 1 - eps], method="brentq"
            ).root
            return ucb

        items_counts = self.items_counts_aggr.withColumn(self.rating_column, get_ucb("pos", "total"))

        self.item_popularity = items_counts.drop("pos", "total")

        self.item_popularity.cache().count()
        self.fill = 1 + math.sqrt(self.coef * math.log(self.full_count))



from typing import Optional

import numpy as np

from replay.data.dataset import Dataset
from replay.utils import PYSPARK_AVAILABLE

from .base_rec import NonPersonalizedRecommender

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf


class ThompsonSampling(NonPersonalizedRecommender):
    """
    Thompson Sampling recommender.

    Bandit model with `efficient exploration-exploitation balance
    <https://proceedings.neurips.cc/paper/2011/file/e53a0a2978c28872a4505bdb51db06dc-Paper.pdf>`_.
    The reward probability of each of the K arms is modeled by a Beta distribution
    which is updated after an arm is selected. The initial prior distribution is Beta(1,1).
    """

    def __init__(
        self,
        sample: bool = False,
        seed: Optional[int] = None,
    ):
        self.sample = sample
        self.seed = seed
        super().__init__(add_cold_items=True, cold_weight=1)

    @property
    def _init_args(self):
        return {"sample": self.sample, "seed": self.seed}

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        self._check_rating(dataset)

        num_positive = (
            dataset.interactions.filter(sf.col(self.rating_column) == sf.lit(1))
            .groupby(self.item_column)
            .agg(sf.count(self.rating_column).alias("positive"))
        )
        num_negative = (
            dataset.interactions.filter(sf.col(self.rating_column) == sf.lit(0))
            .groupby(self.item_column)
            .agg(sf.count(self.rating_column).alias("negative"))
        )

        self.item_popularity = num_positive.join(num_negative, how="inner", on=self.item_column)

        self.item_popularity = self.item_popularity.withColumn(
            self.rating_column, sf.udf(np.random.beta, "double")("positive", "negative")
        ).drop("positive", "negative")
        self.item_popularity.cache().count()
        self.fill = np.random.beta(1, 1)



import abc
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, Literal, Optional

import numpy as np

from replay.utils import TORCH_AVAILABLE, PandasDataFrame

if TORCH_AVAILABLE:
    import torch

MetricName = Literal[
    "recall",
    "precision",
    "ndcg",
    "map",
    "mrr",
    "novelty",
    "coverage",
]

DEFAULT_METRICS: list[MetricName] = [
    "map",
    "ndcg",
    "recall",
]

DEFAULT_KS: list[int] = [1, 5, 10, 20]


@dataclass
class _MetricRequirements:
    """
    Stores description of metrics which need to be computed
    """

    top_k: list[int]
    need_recall: bool
    need_precision: bool
    need_ndcg: bool
    need_map: bool
    need_mrr: bool
    need_novelty: bool
    need_coverage: bool

    def __post_init__(self) -> None:
        metrics = []

        for k in self.top_k:
            if self.need_recall:
                metrics.append(f"recall@{k}")

            if self.need_precision:
                metrics.append(f"precision@{k}")

            if self.need_ndcg:
                metrics.append(f"ndcg@{k}")

            if self.need_map:
                metrics.append(f"map@{k}")

            if self.need_mrr:
                metrics.append(f"mrr@{k}")

            if self.need_novelty:
                metrics.append(f"novelty@{k}")

        self._metric_names = metrics

    @property
    def metric_names(self) -> list[str]:
        """
        Getting metric names
        """
        return self._metric_names

    @classmethod
    def from_metrics(cls, metrics: set[str], top_k: list[int]) -> "_MetricRequirements":
        """
        Creating a class based on a given list of metrics and K values
        """
        return _MetricRequirements(
            top_k=top_k,
            need_recall="recall" in metrics,
            need_precision="precision" in metrics,
            need_ndcg="ndcg" in metrics,
            need_map="map" in metrics,
            need_mrr="mrr" in metrics,
            need_novelty="novelty" in metrics,
            need_coverage="coverage" in metrics,
        )


class _CoverageHelper:
    """
    Computes coverage metric over multiple batches
    """

    def __init__(self, top_k: list[int], item_count: Optional[int]) -> None:
        """
        :param top_k: (list): Consider the highest k scores in the ranking.
        :param item_count: (optional, int): the total number of items in the dataset.
        """
        self._top_k = top_k
        self._item_count = item_count
        self.reset()

    def reset(self) -> None:
        """
        Reload the metric counter
        """
        self._train_hist = torch.zeros(self.item_count)
        self._pred_hist: dict[int, torch.Tensor] = {k: torch.zeros(self.item_count) for k in self._top_k}

    def _ensure_hists_on_device(self, device: torch.device) -> None:
        self._train_hist = self._train_hist.to(device)
        for k in self._top_k:
            self._pred_hist[k] = self._pred_hist[k].to(device)

    def add_prediction(self, predictions: torch.LongTensor) -> None:
        """
        Add a batch with predictions.

        :param predictions: (torch.LongTensor): A batch with the same number of recommendations for each user.
        """
        self._ensure_hists_on_device(predictions.device)
        for k in self._top_k:
            self._pred_hist[k] += torch.histc(
                predictions[:, :k].float(), bins=self.item_count, min=0, max=self.item_count - 1
            )

    def add_train(self, train: torch.LongTensor) -> None:
        """
        Add a training set batch.

        :param train: (optional, int): A batch corresponding to the train set for each user.
            If users have a train set of different sizes then you need to do the padding using -2.
        """
        self._ensure_hists_on_device(train.device)
        flatten_train = train.flatten()
        filtered_train = torch.masked_select(flatten_train, flatten_train != -2)
        self._train_hist += torch.histc(filtered_train.float(), bins=self.item_count, min=0, max=self.item_count - 1)

    def get_metrics(self) -> Mapping[str, float]:
        """
        Getting calculated metrics.
        """
        train_item_count = (self._train_hist > 0).sum().item()
        result = {}
        for k in self._top_k:
            intersection = (self._train_hist > 0) & (self._pred_hist[k] > 0)
            result[f"coverage@{k}"] = intersection.sum().item() / train_item_count
        return result

    @property
    def item_count(self) -> int:
        """
        The number of items in the training dataset
        """
        assert self._item_count
        return self._item_count

    @item_count.setter
    def item_count(self, value: int) -> None:
        self._item_count = value


class _MetricBuilder(abc.ABC):
    """
    Interface of the metrics builder, which is used to combine
    predictions over multiple batches to calculate single result
    """

    @abc.abstractmethod
    def reset(self) -> None:
        """
        Reload the metric counter
        """

    @abc.abstractmethod
    def add_prediction(self, predictions: Any, ground_truth: Any, train: Any = None) -> None:
        """
        Add a batch with predictions
        """

    @abc.abstractmethod
    def get_metrics(self) -> Mapping[str, float]:
        """
        Getting calculated metrics
        """


class TorchMetricsBuilder(_MetricBuilder):
    """
    Computes specified metrics over multiple batches
    """

    def __init__(
        self,
        metrics: list[MetricName] = DEFAULT_METRICS,
        top_k: Optional[list[int]] = DEFAULT_KS,
        item_count: Optional[int] = None,
    ) -> None:
        """
        :param metrics: (list[MetricName]): Names of metrics to calculate.
            Default: `["map", "ndcg", "recall"]`.
        :param top_k: (list): Consider the highest k scores in the ranking.
            Default: `[1, 5, 10, 20]`.
        :param item_count: (optional, int): the total number of items in the dataset.
            You can omit this parameter if you don't need to calculate the Coverage metric.
        """
        self._mr = _MetricRequirements.from_metrics(
            set(metrics),
            sorted(set(top_k)),
        )
        if self._mr.need_ndcg:
            self._ndcg_weights: torch.Tensor
            self._ndcg_idcgs: torch.Tensor
            self._reserve_ndcg_constants()
        if self._mr.need_map:
            self._map_weights: torch.Tensor
            self._reserve_map_constants()
        self._item_count = item_count
        if self._mr.need_coverage:
            assert self._item_count is not None, "For Coverage calculations item_count should be defined."
            self._coverage_helper = _CoverageHelper(top_k=self._mr.top_k, item_count=item_count)
        self.reset()

    @property
    def max_k(self) -> int:
        """
        Maximum K for calculating metrics.
        """
        return max(self._mr.top_k)

    @property
    def item_count(self) -> int:
        """
        The number of items in the training dataset.
        """
        assert self._item_count
        return self._item_count

    @item_count.setter
    def item_count(self, value: int) -> None:
        self._item_count = value

    def reset(self) -> None:
        """
        Reload the metric counter
        """
        self._metric_sum = np.zeros(len(self._mr.metric_names), dtype=np.float64)
        self._prediction_counter = 0
        if self._mr.need_coverage:
            self._coverage_helper.item_count = self.item_count
            self._coverage_helper.reset()

    def _ensure_constants_on_device(self, device: torch.device) -> None:
        if self._mr.need_ndcg:
            self._ndcg_weights = self._ndcg_weights.to(device)
            self._ndcg_idcgs = self._ndcg_idcgs.to(device)
        if self._mr.need_map:
            self._map_weights = self._map_weights.to(device)

    def add_prediction(
        self,
        predictions: torch.LongTensor,
        ground_truth: torch.LongTensor,
        train: Optional[torch.LongTensor] = None,
    ) -> None:
        """
        Add a batch with predictions, ground truth and train set to calculate the metrics.

        :param predictions: (torch.LongTensor): A batch with the same number of recommendations for each user.
        :param ground_truth: (torch.LongTensor): A batch corresponding to the test set for each user.
            If users have a test set of different sizes then you need to do the padding using -1.
        :param train: (optional, int): A batch corresponding to the train set for each user.
            If users have a train set of different sizes then you need to do the padding using -2.
            You can omit this parameter if you don't need to calculate the coverage or novelty metrics.
        """
        self._ensure_constants_on_device(predictions.device)
        metrics_sum = np.array(self._compute_metrics_sum(predictions, ground_truth, train), dtype=np.float64)
        if self._mr.need_coverage:
            self._coverage_helper.add_prediction(predictions)
            assert train is not None
            self._coverage_helper.add_train(train)
        self._prediction_counter += len(predictions)
        self._metric_sum += metrics_sum

    def get_metrics(self) -> Mapping[str, float]:
        """
        Getting calculated metrics.
        """
        assert self._prediction_counter > 0
        metrics = self._metric_sum / self._prediction_counter
        result = dict(zip(self._mr.metric_names, metrics))
        if self._mr.need_coverage:
            result.update(self._coverage_helper.get_metrics())
        return result

    def _compute_recall(self, hits: torch.Tensor, ground_truth_count: torch.Tensor) -> float:
        recall = hits.sum(1) / ground_truth_count
        return recall.sum().item()

    def _compute_precision(self, hits: torch.Tensor, k: int) -> float:
        precision = hits.sum(1) / k
        return precision.sum().item()

    def _compute_ndcg(self, hits: torch.Tensor, ground_truth_count_at_k: torch.Tensor, k: int) -> float:
        dcg = (hits * self._ndcg_weights[:k]).sum(1)
        idcg = self._ndcg_idcgs[ground_truth_count_at_k]
        ndcg = dcg / idcg
        return ndcg.sum().item()

    def _compute_map(self, hits: torch.Tensor, ground_truth_count_at_k: torch.Tensor, k: int) -> float:
        hits_cumsum = hits * hits.cumsum(1)
        hits_cumsum_discounted = (hits_cumsum * self._map_weights[:k]).sum(1)
        average_precision = hits_cumsum_discounted / ground_truth_count_at_k
        return average_precision.sum().item()

    def _compute_mrr(self, hits: torch.Tensor) -> float:
        indexed_hits = hits * torch.arange(hits.shape[1], 0, -1, device=hits.device)
        vals, indices = torch.max(indexed_hits, dim=1)
        fixed_indices = indices.masked_fill(vals == 0, -2) + 1
        mrr = (1 / fixed_indices).clamp(min=0)
        return mrr.sum().item()

    def _compute_novelty(self, train_hits: torch.Tensor, k: int) -> float:
        novelty = (~train_hits).sum(1) / k
        return novelty.sum().item()

    def _compute_metrics_sum(
        self, predictions: torch.LongTensor, ground_truth: torch.LongTensor, train: Optional[torch.LongTensor]
    ) -> list[float]:
        result: list[float] = []

        # Getting a tensor of the same size as predictions
        # The tensor contains information about whether the item from the prediction is present in the test set
        item_hits = (predictions.unsqueeze(1) == ground_truth.unsqueeze(-1)).any(dim=1)
        if self._mr.need_novelty:
            assert train is not None
            train_hits = (predictions.unsqueeze(1) == train.unsqueeze(-1)).any(dim=1)

        hits = item_hits[:, : self.max_k].float()

        # must be > 0, otherwise it may lead to NaNs in metrics
        gt_count = (ground_truth >= 0).sum(1).clamp(min=1)

        for k in self._mr.top_k:
            hits_at_k = hits[:, :k]
            gt_count_at_k = gt_count.clamp(max=k)

            if self._mr.need_recall:
                result.append(self._compute_recall(hits_at_k, gt_count))

            if self._mr.need_precision:
                result.append(self._compute_precision(hits_at_k, k))

            if self._mr.need_ndcg:
                result.append(self._compute_ndcg(hits_at_k, gt_count_at_k, k))

            if self._mr.need_map:
                result.append(self._compute_map(hits_at_k, gt_count_at_k, k))

            if self._mr.need_mrr:
                result.append(self._compute_mrr(hits_at_k))

            if self._mr.need_novelty:
                train_hits_at_k = train_hits[:, :k]
                result.append(self._compute_novelty(train_hits_at_k, k))

        return result

    def _reserve_ndcg_constants(self) -> None:
        position = torch.arange(2, 2 + self.max_k).float()
        self._ndcg_weights = 1 / torch.log2(position)
        self._ndcg_idcgs = torch.cat(
            [
                torch.tensor([0], dtype=torch.float32),
                self._ndcg_weights.cumsum(dim=0),
            ]
        )

    def _reserve_map_constants(self) -> None:
        self._map_weights = 1.0 / torch.arange(1, 1 + self.max_k).float()


def metrics_to_df(metrics: Mapping[str, float]) -> PandasDataFrame:
    """
    Converting metrics to Pandas DataFrame
    """
    metrics_df = PandasDataFrame(metrics.items(), columns=["m", "v"])

    metric_name_and_k = metrics_df["m"].str.split("@", expand=True)
    metrics_df["metric"] = metric_name_and_k[0]
    metrics_df["k"] = metric_name_and_k[1]

    pivoted_metrics = metrics_df.pivot(index="metric", columns="k", values="v")
    pivoted_metrics.index.name = None

    pivoted_metrics.sort_index(axis=0, inplace=True)
    pivoted_metrics.sort_index(axis=1, inplace=True)

    return pivoted_metrics



from typing import TYPE_CHECKING

from replay.utils import PandasDataFrame, PolarsDataFrame, SparkDataFrame

from .base_metric import Metric, MetricsDataFrameLike, MetricsReturnType

if TYPE_CHECKING:  # pragma: no cover
    __class__: type


class Novelty(Metric):
    """
    Measure the fraction of shown items in recommendation list, that users\
        didn't see in train dataset.

    .. math::
        Novelty@K(i) = \\frac
        {\\parallel {R^{i}_{1..\\min(K, \\parallel R^{i} \\parallel)} \\setminus train^{i}} \\parallel}
        {K}

    .. math::
        Novelty@K = \\frac {1}{N}\\sum_{i=1}^{N}Novelty@K(i)

    :math:`R^{i}` -- the recommendations for the :math:`i`-th user.

    :math:`R^{i}_{j}` -- the :math:`j`-th recommended item for the :math:`i`-th user.

    :math:`R_{1..j}^{i}` -- the first :math:`j` recommendations for the :math:`i`-th user.

    :math:`train^{i}` -- the train items of the :math:`i`-th user.

    :math:`N` -- the number of users.

    :Based on:

        P. Castells, S. Vargas, and J. Wang, Novelty and diversity metrics for recommender systems:
        choice, discovery and relevance, ECIR 2011.
        `Link <https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=ecb23352b3fd8abd32332790fda7aca59c498fdf>`_.

    >>> recommendations
       query_id  item_id  rating
    0         1        3    0.6
    1         1        7    0.5
    2         1       10    0.4
    3         1       11    0.3
    4         1        2    0.2
    5         2        5    0.6
    6         2        8    0.5
    7         2       11    0.4
    8         2        1    0.3
    9         2        3    0.2
    10        3        4    1.0
    11        3        9    0.5
    12        3        2    0.1
    >>> train
       query_id  item_id
    0         1        5
    1         1        6
    2         1        8
    3         1        9
    4         1        2
    5         2        5
    6         2        8
    7         2       11
    8         2        1
    9         2        3
    10        3        4
    11        3        9
    12        3        2
    >>> from replay.metrics import Median, ConfidenceInterval, PerUser
    >>> Novelty(2)(recommendations, train)
    {'Novelty@2': 0.3333333333333333}
    >>> Novelty(2, mode=PerUser())(recommendations, train)
    {'Novelty-PerUser@2': {1: 1.0, 2: 0.0, 3: 0.0}}
    >>> Novelty(2, mode=Median())(recommendations, train)
    {'Novelty-Median@2': 0.0}
    >>> Novelty(2, mode=ConfidenceInterval(alpha=0.95))(recommendations, train)
    {'Novelty-ConfidenceInterval@2': 0.6533213281800181}
    <BLANKLINE>
    """

    def __call__(
        self,
        recommendations: MetricsDataFrameLike,
        train: MetricsDataFrameLike,
    ) -> MetricsReturnType:
        """
        Compute metric.

        :param recommendations: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict):
            model predictions.
            If DataFrame then it must contains user, item and score columns.
            If dict then items must be sorted in decreasing order of their scores.
        :param train: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict, optional):
            train data.
            If DataFrame then it must contains user and item columns.

        :return: metric values
        """
        self._check_dataframes_equal_types(recommendations, train)
        if isinstance(recommendations, SparkDataFrame):
            self._check_duplicates_spark(recommendations)
            assert isinstance(train, SparkDataFrame)
            return self._spark_call(recommendations, train)
        if isinstance(recommendations, PolarsDataFrame):
            self._check_duplicates_polars(recommendations)
            assert isinstance(train, PolarsDataFrame)
            return self._polars_call(recommendations, train)
        is_pandas = isinstance(recommendations, PandasDataFrame)
        recommendations = (
            self._convert_pandas_to_dict_with_score(recommendations)
            if is_pandas
            else self._convert_dict_to_dict_with_score(recommendations)
        )
        self._check_duplicates_dict(recommendations)
        train = self._convert_pandas_to_dict_without_score(train) if is_pandas else train
        assert isinstance(train, dict)

        return self._dict_call(
            list(train),
            pred_item_id=recommendations,
            train=train,
        )

    def _spark_call(self, recommendations: SparkDataFrame, train: SparkDataFrame) -> MetricsReturnType:
        """
        Implementation for Pyspark DataFrame.
        """
        recs = self._get_enriched_recommendations(recommendations, train).withColumnRenamed("ground_truth", "train")
        recs = self._rearrange_columns(recs)
        return self._spark_compute(recs)

    def _polars_call(self, recommendations: PolarsDataFrame, train: PolarsDataFrame) -> MetricsReturnType:
        """
        Implementation for Polars DataFrame.
        """
        recs = self._get_enriched_recommendations(recommendations, train).rename({"ground_truth": "train"})
        recs = self._rearrange_columns(recs)
        return self._polars_compute(recs)

    @staticmethod
    def _get_metric_value_by_user(ks: list[int], pred: list, train: list) -> list[float]:
        if not train or not pred:
            return [1.0 for _ in ks]
        set_train = set(train)
        return [1.0 - len(set(pred[:k]) & set_train) / len(pred[:k]) for k in ks]



import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

import polars as pl

from replay.utils import (
    PYSPARK_AVAILABLE,
    DataFrameLike,
    PandasDataFrame,
    PolarsDataFrame,
    SparkDataFrame,
)

if PYSPARK_AVAILABLE:
    from pyspark.sql import (
        Window,
        functions as sf,
    )


SplitterReturnType = tuple[DataFrameLike, DataFrameLike]


class Splitter(ABC):
    """Base class"""

    _init_arg_names = [
        "drop_cold_users",
        "drop_cold_items",
        "query_column",
        "item_column",
        "timestamp_column",
        "session_id_column",
        "session_id_processing_strategy",
    ]

    def __init__(
        self,
        drop_cold_items: bool = False,
        drop_cold_users: bool = False,
        query_column: str = "query_id",
        item_column: Optional[str] = "item_id",
        timestamp_column: Optional[str] = "timestamp",
        session_id_column: Optional[str] = None,
        session_id_processing_strategy: str = "test",
    ):
        """
        :param drop_cold_items: flag to remove items that are not in train data
        :param drop_cold_users: flag to remove users that are not in train data
        :param query_column: query id column name
        :param item_column: item id column name
        :param timestamp_column: timestamp column name
        :param session_id_column: name of session id column, which values can not be split.
        :param session_id_processing_strategy: strategy of processing session if it is split,
            values: ``train, test``, train: whole split session goes to train. test: same but to test.
            default: ``test``.
        """
        self.drop_cold_users = drop_cold_users
        self.drop_cold_items = drop_cold_items
        self.query_column = query_column
        self.item_column = item_column
        self.timestamp_column = timestamp_column

        self.session_id_column = session_id_column
        self.session_id_processing_strategy = session_id_processing_strategy

    @property
    def _init_args(self):
        return {name: getattr(self, name) for name in self._init_arg_names}

    def save(self, path: str) -> None:
        """
        Method for saving splitter in `.replay` directory.
        """
        base_path = Path(path).with_suffix(".replay").resolve()
        base_path.mkdir(parents=True, exist_ok=True)

        splitter_dict = {}
        splitter_dict["init_args"] = self._init_args
        splitter_dict["_class_name"] = str(self)

        with open(base_path / "init_args.json", "w+") as file:
            json.dump(splitter_dict, file)

    @classmethod
    def load(cls, path: str, **kwargs) -> "Splitter":
        """
        Method for loading splitter from `.replay` directory.
        """
        base_path = Path(path).with_suffix(".replay").resolve()
        with open(base_path / "init_args.json") as file:
            splitter_dict = json.loads(file.read())
        splitter = cls(**splitter_dict["init_args"])

        return splitter

    def __str__(self):
        return type(self).__name__

    def _drop_cold_items_and_users(
        self,
        train: DataFrameLike,
        test: DataFrameLike,
    ) -> DataFrameLike:
        if isinstance(train, type(test)) is False:
            msg = "Train and test dataframes must have consistent types"
            raise TypeError(msg)

        if isinstance(test, SparkDataFrame):
            return self._drop_cold_items_and_users_from_spark(train, test)
        if isinstance(test, PandasDataFrame):
            return self._drop_cold_items_and_users_from_pandas(train, test)
        else:
            return self._drop_cold_items_and_users_from_polars(train, test)

    def _drop_cold_items_and_users_from_pandas(
        self,
        train: PandasDataFrame,
        test: PandasDataFrame,
    ) -> PandasDataFrame:
        if self.drop_cold_items:
            test = test[test[self.item_column].isin(train[self.item_column])]

        if self.drop_cold_users:
            test = test[test[self.query_column].isin(train[self.query_column])]

        return test

    def _drop_cold_items_and_users_from_spark(
        self,
        train: SparkDataFrame,
        test: SparkDataFrame,
    ) -> SparkDataFrame:
        if self.drop_cold_items:
            train_tmp = train.select(sf.col(self.item_column).alias("item")).distinct()
            test = test.join(train_tmp, train_tmp["item"] == test[self.item_column]).drop("item")

        if self.drop_cold_users:
            train_tmp = train.select(sf.col(self.query_column).alias("user")).distinct()
            test = test.join(train_tmp, train_tmp["user"] == test[self.query_column]).drop("user")

        return test

    def _drop_cold_items_and_users_from_polars(
        self,
        train: PolarsDataFrame,
        test: PolarsDataFrame,
    ) -> PolarsDataFrame:
        if self.drop_cold_items:
            train_tmp = train.select(self.item_column).unique()
            test = test.join(train_tmp, on=self.item_column)

        if self.drop_cold_users:
            train_tmp = train.select(self.query_column).unique()
            test = test.join(train_tmp, on=self.query_column)

        return test

    @abstractmethod
    def _core_split(self, interactions: DataFrameLike) -> SplitterReturnType:
        """
        This method implements split strategy

        :param interactions: input DataFrame `[timestamp, user_id, item_id, relevance]`
        :returns: `train` and `test DataFrames
        """

    def split(self, interactions: DataFrameLike) -> SplitterReturnType:
        """
        Splits input DataFrame into train and test

        :param interactions: input DataFrame ``[timestamp, user_id, item_id, relevance]``
        :returns: List of splitted DataFrames
        """
        train, test = self._core_split(interactions)
        test = self._drop_cold_items_and_users(train, test)

        return train, test

    def _recalculate_with_session_id_column(self, data: DataFrameLike) -> DataFrameLike:
        if isinstance(data, SparkDataFrame):
            return self._recalculate_with_session_id_column_spark(data)
        if isinstance(data, PandasDataFrame):
            return self._recalculate_with_session_id_column_pandas(data)
        else:
            return self._recalculate_with_session_id_column_polars(data)

    def _recalculate_with_session_id_column_pandas(self, data: PandasDataFrame) -> PandasDataFrame:
        agg_function_name = "first" if self.session_id_processing_strategy == "train" else "last"
        res = data.copy()
        res["is_test"] = res.groupby([self.query_column, self.session_id_column])["is_test"].transform(
            agg_function_name
        )

        return res

    def _recalculate_with_session_id_column_spark(self, data: SparkDataFrame) -> SparkDataFrame:
        agg_function = sf.first if self.session_id_processing_strategy == "train" else sf.last
        res = data.withColumn(
            "is_test",
            agg_function("is_test").over(
                Window.orderBy(self.timestamp_column)
                .partitionBy(self.query_column, self.session_id_column)
                .rowsBetween(Window.unboundedPreceding, Window.unboundedFollowing)
            ),
        )

        return res

    def _recalculate_with_session_id_column_polars(self, data: PolarsDataFrame) -> PolarsDataFrame:
        agg_function = pl.Expr.first if self.session_id_processing_strategy == "train" else pl.Expr.last
        res = data.with_columns(
            agg_function(pl.col("is_test").sort_by(self.timestamp_column)).over(
                [self.query_column, self.session_id_column]
            )
        )

        return res



import abc
import json
from pathlib import Path
from typing import TYPE_CHECKING, Union

import numpy as np
import pandas as pd
import polars as pl
from pandas import DataFrame as PandasDataFrame
from polars import DataFrame as PolarsDataFrame

if TYPE_CHECKING:
    from .schema import TensorSchema


class SequentialDataset(abc.ABC):
    """
    Abstract base class for sequential dataset
    """

    @abc.abstractmethod
    def __len__(self) -> int:  # pragma: no cover
        """
        Returns the length of the dataset.
        """

    @abc.abstractmethod
    def get_query_id(self, index: int) -> int:  # pragma: no cover
        """
        Getting a query id for a given index.

        :param index: the row number in the dataset.
        """

    @abc.abstractmethod
    def get_all_query_ids(self) -> np.ndarray:  # pragma: no cover
        """
        Getting a list of all query ids.
        """

    @abc.abstractmethod
    def get_sequence_length(self, index: int) -> int:  # pragma: no cover
        """
        Returns the length of the sequence at the specified index.

        :param index: the row number in the dataset.
        """

    @abc.abstractmethod
    def get_max_sequence_length(self) -> int:  # pragma: no cover
        """
        Returns the maximum length among all sequences from the `SequentialDataset`.
        """

    @abc.abstractmethod
    def get_sequence(self, index: Union[int, np.ndarray], feature_name: str) -> np.ndarray:  # pragma: no cover
        """
        Getting a sequence based on a given index and feature name.

        :param index: single index or list of indices.
        :param feature_name: the name of the feature.
        """

    @abc.abstractmethod
    def get_sequence_by_query_id(
        self, query_id: Union[int, np.ndarray], feature_name: str
    ) -> np.ndarray:  # pragma: no cover
        """
        Getting a sequence based on a given query id and feature name.

        :param query_id: single query id or list of query ids.
        :param feature_name: the name of the feature.
        """

    @abc.abstractmethod
    def filter_by_query_id(self, query_ids_to_keep: np.ndarray) -> "SequentialDataset":  # pragma: no cover
        """
        Returns a `SequentialDataset` that contains only query ids from the specified list.

        :param query_ids_to_keep: list of query ids.
        """

    @property
    @abc.abstractmethod
    def schema(self) -> "TensorSchema":  # pragma: no cover
        """
        :returns: List of tensor features.
        """

    @staticmethod
    def keep_common_query_ids(
        lhs: "SequentialDataset", rhs: "SequentialDataset"
    ) -> tuple["SequentialDataset", "SequentialDataset"]:
        """
        Returns `SequentialDatasets` that contain query ids from both datasets.

        :param lhs: `SequentialDataset`.
        :param rhs: `SequentialDataset`.
        """
        lhs_queries = lhs.get_all_query_ids()
        rhs_queries = rhs.get_all_query_ids()
        common_queries = np.intersect1d(lhs_queries, rhs_queries, assume_unique=True)
        lhs_filtered = lhs.filter_by_query_id(common_queries)
        rhs_filtered = rhs.filter_by_query_id(common_queries)
        return lhs_filtered, rhs_filtered

    def save(self, path: str) -> None:
        base_path = Path(path).with_suffix(".replay").resolve()
        base_path.mkdir(parents=True, exist_ok=True)

        sequential_dict = {}
        sequential_dict["_class_name"] = self.__class__.__name__

        df = SequentialDataset._convert_array_to_list(self._sequences)
        df.reset_index().to_parquet(base_path / "sequences.parquet")
        sequential_dict["init_args"] = {
            "tensor_schema": self._tensor_schema._get_object_args(),
            "query_id_column": self._query_id_column,
            "item_id_column": self._item_id_column,
            "sequences_path": "sequences.parquet",
        }

        with open(base_path / "init_args.json", "w+") as file:
            json.dump(sequential_dict, file)

    @staticmethod
    def _convert_array_to_list(df):
        return df.map(lambda x: x.tolist() if isinstance(x, np.ndarray) else x)

    @staticmethod
    def _convert_list_to_array(df):
        return df.map(lambda x: np.array(x) if isinstance(x, list) else x)


class PandasSequentialDataset(SequentialDataset):
    """
    Sequential dataset that stores sequences in PandasDataFrame format.
    """

    def __init__(
        self,
        tensor_schema: "TensorSchema",
        query_id_column: str,
        item_id_column: str,
        sequences: PandasDataFrame,
    ) -> None:
        """
        :param tensor_schema: schema of tensor features.
        :param query_id_column: The name of the column containing query ids.
        :param item_id_column: The name of the column containing item ids.
        :param sequences: PandasDataFrame with sequences corresponding to the tensor schema.
        """
        self._check_if_schema_matches_data(tensor_schema, sequences)

        self._tensor_schema = tensor_schema
        self._query_id_column = query_id_column
        self._item_id_column = item_id_column

        if sequences.index.name != query_id_column:
            sequences = sequences.set_index(query_id_column)

        self._sequences = SequentialDataset._convert_list_to_array(sequences)

    def __len__(self) -> int:
        return len(self._sequences)

    def get_query_id(self, index: int) -> int:
        return self._sequences.index[index]

    def get_all_query_ids(self) -> np.ndarray:
        return self._sequences.index.values

    def get_sequence_length(self, index: int) -> int:
        return len(self._sequences[self._item_id_column].iloc[index])

    def get_max_sequence_length(self) -> int:
        return max(len(seq) for seq in self._sequences[self._item_id_column])

    def get_sequence(self, index: Union[int, np.ndarray], feature_name: str) -> np.ndarray:
        return np.array(self._sequences[feature_name].iloc[index])

    def get_sequence_by_query_id(self, query_id: Union[int, np.ndarray], feature_name: str) -> np.ndarray:
        try:
            return np.array(self._sequences[feature_name].loc[query_id])
        except KeyError:
            return np.array([], dtype=np.int64)

    def filter_by_query_id(self, query_ids_to_keep: np.ndarray) -> "PandasSequentialDataset":
        filtered_sequences = self._sequences.loc[query_ids_to_keep]
        return PandasSequentialDataset(
            tensor_schema=self._tensor_schema,
            query_id_column=self._query_id_column,
            item_id_column=self._item_id_column,
            sequences=filtered_sequences,
        )

    @property
    def schema(self) -> "TensorSchema":
        return self._tensor_schema

    @classmethod
    def _check_if_schema_matches_data(cls, tensor_schema: "TensorSchema", data: PandasDataFrame) -> None:
        for tensor_feature_name in tensor_schema:
            if tensor_feature_name not in data:
                msg = "Tensor schema does not match with provided data frame"
                raise ValueError(msg)

    @classmethod
    def load(cls, path: str, **kwargs) -> "PandasSequentialDataset":
        """
        Method for loading PandasSequentialDataset object from `.replay` directory.
        """
        from replay.data.nn import TensorSchema

        base_path = Path(path).with_suffix(".replay").resolve()
        with open(base_path / "init_args.json") as file:
            sequential_dict = json.loads(file.read())

        sequences = pd.read_parquet(base_path / sequential_dict["init_args"]["sequences_path"])
        sequences = cls._convert_array_to_list(sequences)
        dataset = cls(
            tensor_schema=TensorSchema._create_object_by_args(sequential_dict["init_args"]["tensor_schema"]),
            query_id_column=sequential_dict["init_args"]["query_id_column"],
            item_id_column=sequential_dict["init_args"]["item_id_column"],
            sequences=sequences,
        )

        return dataset


class PolarsSequentialDataset(PandasSequentialDataset):
    """
    Sequential dataset that stores sequences in PolarsDataFrame format.
    """

    def __init__(
        self,
        tensor_schema: "TensorSchema",
        query_id_column: str,
        item_id_column: str,
        sequences: PolarsDataFrame,
    ) -> None:
        """
        :param tensor_schema: schema of tensor features.
        :param query_id_column: The name of the column containing query ids.
        :param item_id_column: The name of the column containing item ids.
        :param sequences: PolarsDataFrame with sequences corresponding to the tensor schema.
        """
        self._check_if_schema_matches_data(tensor_schema, sequences)

        self._tensor_schema = tensor_schema
        self._query_id_column = query_id_column
        self._item_id_column = item_id_column

        self._sequences = self._convert_polars_to_pandas(sequences)
        if self._sequences.index.name != query_id_column:
            self._sequences = self._sequences.set_index(query_id_column)

    def filter_by_query_id(self, query_ids_to_keep: np.ndarray) -> "PolarsSequentialDataset":
        filtered_sequences = self._sequences.loc[query_ids_to_keep]
        if filtered_sequences.index.name == self._query_id_column:
            filtered_sequences = filtered_sequences.reset_index()
        return PolarsSequentialDataset(
            tensor_schema=self._tensor_schema,
            query_id_column=self._query_id_column,
            item_id_column=self._item_id_column,
            sequences=self._convert_pandas_to_polars(filtered_sequences),
        )

    def _convert_polars_to_pandas(self, df: PolarsDataFrame) -> PandasDataFrame:
        pandas_df = PandasDataFrame(df.to_dict(as_series=False))
        pandas_df = SequentialDataset._convert_list_to_array(pandas_df)
        return pandas_df

    def _convert_pandas_to_polars(self, df: PandasDataFrame) -> PolarsDataFrame:
        df = SequentialDataset._convert_array_to_list(df)
        return pl.from_dict(df.to_dict("list"))

    @classmethod
    def _check_if_schema_matches_data(cls, tensor_schema: "TensorSchema", data: PolarsDataFrame) -> None:
        for tensor_feature_name in tensor_schema:
            if tensor_feature_name not in data:
                msg = "Tensor schema does not match with provided data frame"
                raise ValueError(msg)

    @classmethod
    def load(cls, path: str, **kwargs) -> "PandasSequentialDataset":
        """
        Method for loading PandasSequentialDataset object from `.replay` directory.
        """
        from replay.data.nn import TensorSchema

        base_path = Path(path).with_suffix(".replay").resolve()
        with open(base_path / "init_args.json") as file:
            sequential_dict = json.loads(file.read())

        sequences = pl.from_pandas(pd.read_parquet(base_path / sequential_dict["init_args"]["sequences_path"]))
        dataset = cls(
            tensor_schema=TensorSchema._create_object_by_args(sequential_dict["init_args"]["tensor_schema"]),
            query_id_column=sequential_dict["init_args"]["query_id_column"],
            item_id_column=sequential_dict["init_args"]["item_id_column"],
            sequences=sequences,
        )

        return dataset



from typing import Optional

import numpy as np
from scipy.sparse import csr_matrix

from replay.utils import PYSPARK_AVAILABLE, DataFrameLike, SparkDataFrame

if PYSPARK_AVAILABLE:
    from replay.utils.spark_utils import spark_to_pandas


class CSRConverter:
    """
    Convert input data to csr sparse matrix.
    Where ``data_column``, ``first_dim_column`` and ``second_dim_column`` satisfy the relationship
    ``matrix[first_dim_column[i], second_dim_column[i]] = data_column[i]``.

    >>> import pandas as pd
    >>> interactions = pd.DataFrame({
    ...    "user_id": [1, 1, 1, 2, 2, 2, 3, 3, 3, 3],
    ...    "item_id": [3, 7, 10, 5, 8, 11, 4, 9, 2, 5],
    ...    "rating": [1, 2, 3, 3, 2, 1, 3, 12, 1, 4]
    ... })
    >>> interactions
        user_id  item_id  rating
    0        1        3       1
    1        1        7       2
    2        1       10       3
    3        2        5       3
    4        2        8       2
    5        2       11       1
    6        3        4       3
    7        3        9      12
    8        3        2       1
    9        3        5       4
    >>> csr_interactions = CSRConverter(
    ...    first_dim_column="user_id",
    ...    second_dim_column="item_id",
    ...    data_column="rating",
    ... ).transform(interactions)
    >>> csr_interactions.todense()
    matrix([[ 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [ 0,  0,  0,  1,  0,  0,  0,  2,  0,  0,  3,  0],
            [ 0,  0,  0,  0,  0,  3,  0,  0,  2,  0,  0,  1],
            [ 0,  0,  1,  0,  3,  4,  0,  0,  0, 12,  0,  0]])
    <BLANKLINE>
    """

    def __init__(
        self,
        first_dim_column: str,
        second_dim_column: str,
        data_column: Optional[str] = None,
        row_count: Optional[int] = None,
        column_count: Optional[int] = None,
        allow_collect_to_master: bool = False,
    ):
        """
        :param first_dim_column: The name of the column for first dimension in resulting sparse matrix,
            default: ``user_id``.
        :param second_dim_column: The name of the column for second dimension in resulting sparse matrix,
            default: ``item_id``.
        :param data_column: The name of the column with values at the intersection
            of the corresponding row and column. If ``None`` then ones will be taken. Default: ``None``.
        :param row_count: Number of rows in resulting sparse matrix.
            If ``None`` then it depends on the maximum value in first_dim_column. Default: ``None``.
        :param column_count: Number of columns for resulting sparse matrix.
            If ``None`` then it depends on the maximum value in second_dim_column. Default: ``None``.
        :param allow_collect_to_master: Flag allowing spark
            to make a collection to the master node, default: ``False``.
        """
        self.first_dim_column = first_dim_column
        self.second_dim_column = second_dim_column
        self.data_column = data_column
        self.row_count = row_count
        self.column_count = column_count
        self.allow_collect_to_master = allow_collect_to_master

    def transform(self, data: DataFrameLike) -> csr_matrix:
        """
        Transform Spark or Pandas Data Frame to csr.

        :param data: Spark or Pandas Data Frame containing columns
            ``first_dim_column``, ``second_dim_column``, and optional ``data_column``.

        :returns: Sparse interactions.
        """

        if isinstance(data, SparkDataFrame):
            cols = [self.first_dim_column, self.second_dim_column]
            if self.data_column is not None:
                cols.append(self.data_column)
            data = spark_to_pandas(data.select(cols), self.allow_collect_to_master)

        rows_data = data[self.first_dim_column].values
        cols_data = data[self.second_dim_column].values
        data = data[self.data_column].values if self.data_column is not None else np.ones(data.shape[0])

        def _get_max(data: np.ndarray) -> int:
            return np.max(data) if data.shape[0] > 0 else 0

        row_count = self.row_count if self.row_count is not None else _get_max(rows_data) + 1
        col_count = self.column_count if self.column_count is not None else _get_max(cols_data) + 1
        return csr_matrix(
            (data.tolist(), (rows_data.tolist(), cols_data.tolist())),
            shape=(row_count, col_count),
        )



from datetime import datetime
from typing import Optional, Union

import polars as pl

from replay.utils import (
    PYSPARK_AVAILABLE,
    DataFrameLike,
    PandasDataFrame,
    PolarsDataFrame,
    SparkDataFrame,
)

from .base_splitter import Splitter

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf
    from pyspark.sql import Window


class TimeSplitter(Splitter):
    """
    Split interactions by time.

    >>> from datetime import datetime
    >>> import pandas as pd
    >>> columns = ["query_id", "item_id", "timestamp"]
    >>> data = [
    ...     (1, 1, "01-01-2020"),
    ...     (1, 2, "02-01-2020"),
    ...     (1, 3, "03-01-2020"),
    ...     (1, 4, "04-01-2020"),
    ...     (1, 5, "05-01-2020"),
    ...     (2, 1, "06-01-2020"),
    ...     (2, 2, "07-01-2020"),
    ...     (2, 3, "08-01-2020"),
    ...     (2, 9, "09-01-2020"),
    ...     (2, 10, "10-01-2020"),
    ...     (3, 1, "01-01-2020"),
    ...     (3, 5, "02-01-2020"),
    ...     (3, 3, "03-01-2020"),
    ...     (3, 1, "04-01-2020"),
    ...     (3, 2, "05-01-2020"),
    ... ]
    >>> dataset = pd.DataFrame(data, columns=columns)
    >>> dataset["timestamp"] = pd.to_datetime(dataset["timestamp"], format="%d-%m-%Y")
    >>> dataset
       query_id  item_id  timestamp
    0         1        1 2020-01-01
    1         1        2 2020-01-02
    2         1        3 2020-01-03
    3         1        4 2020-01-04
    4         1        5 2020-01-05
    5         2        1 2020-01-06
    6         2        2 2020-01-07
    7         2        3 2020-01-08
    8         2        9 2020-01-09
    9         2       10 2020-01-10
    10        3        1 2020-01-01
    11        3        5 2020-01-02
    12        3        3 2020-01-03
    13        3        1 2020-01-04
    14        3        2 2020-01-05
    >>> train, test = TimeSplitter(
    ...     time_threshold=datetime.strptime("2020-01-04", "%Y-%M-%d")
    ... ).split(dataset)
    >>> train
       query_id  item_id  timestamp
    0         1        1 2020-01-01
    1         1        2 2020-01-02
    2         1        3 2020-01-03
    3         1        4 2020-01-04
    10        3        1 2020-01-01
    11        3        5 2020-01-02
    12        3        3 2020-01-03
    13        3        1 2020-01-04
    >>> test
       query_id  item_id  timestamp
    4         1        5 2020-01-05
    5         2        1 2020-01-06
    6         2        2 2020-01-07
    7         2        3 2020-01-08
    8         2        9 2020-01-09
    9         2       10 2020-01-10
    14        3        2 2020-01-05
    <BLANKLINE>
    """

    _init_arg_names = [
        "time_threshold",
        "drop_cold_users",
        "drop_cold_items",
        "query_column",
        "item_column",
        "timestamp_column",
        "session_id_column",
        "session_id_processing_strategy",
        "time_column_format",
    ]

    def __init__(
        self,
        time_threshold: Union[datetime, str, float],
        query_column: str = "query_id",
        drop_cold_users: bool = False,
        drop_cold_items: bool = False,
        item_column: str = "item_id",
        timestamp_column: str = "timestamp",
        session_id_column: Optional[str] = None,
        session_id_processing_strategy: str = "test",
        time_column_format: str = "%Y-%m-%d %H:%M:%S",
    ):
        """
        :param time_threshold: Test threshold, can be datetime, string, int or float.
            datetime is in case of splitting by datetime,
            int      is in case of splitting by datetime (Unix format),
            string   will be converted to datetime using ``time_column_format``,
            float    is in case of splitting by ratio, the value must be between 0 and 1.
        :param query_column: Name of user interaction column.
        :param drop_cold_users: Drop users from test DataFrame.
            which are not in train DataFrame, default: False.
        :param drop_cold_items: Drop items from test DataFrame
            which are not in train DataFrame, default: False.
        :param item_column: Name of item interaction column.
            If ``drop_cold_items`` is ``False``, then you can omit this parameter.
            Default: ``item_id``.
        :param timestamp_column: Name of time column,
            Default: ``timestamp``.
        :param session_id_column: Name of session id column, which values can not be split,
            default: ``None``.
        :param session_id_processing_strategy: strategy of processing session if it is split,
            values: ``train, test``, train: whole split session goes to train. test: same but to test.
            default: ``test``.
        """
        super().__init__(
            drop_cold_users=drop_cold_users,
            drop_cold_items=drop_cold_items,
            query_column=query_column,
            item_column=item_column,
            timestamp_column=timestamp_column,
            session_id_column=session_id_column,
            session_id_processing_strategy=session_id_processing_strategy,
        )
        self._precision = 3
        self.time_column_format = time_column_format
        if isinstance(time_threshold, float) and (time_threshold < 0 or time_threshold > 1):
            msg = "time_threshold must be between 0 and 1"
            raise ValueError(msg)
        self.time_threshold = time_threshold

    def _partial_split(
        self, interactions: DataFrameLike, threshold: Union[datetime, str, int]
    ) -> tuple[DataFrameLike, DataFrameLike]:
        if isinstance(threshold, str):
            threshold = datetime.strptime(threshold, self.time_column_format)

        if isinstance(interactions, SparkDataFrame):
            return self._partial_split_spark(interactions, threshold)
        if isinstance(interactions, PandasDataFrame):
            return self._partial_split_pandas(interactions, threshold)
        if isinstance(interactions, PolarsDataFrame):
            return self._partial_split_polars(interactions, threshold)

        msg = f"{self} is not implemented for {type(interactions)}"
        raise NotImplementedError(msg)

    def _partial_split_pandas(
        self, interactions: PandasDataFrame, threshold: Union[datetime, str, int]
    ) -> tuple[PandasDataFrame, PandasDataFrame]:
        res = interactions.copy(deep=True)
        if isinstance(threshold, float):
            res.sort_values(self.timestamp_column, inplace=True)
            test_start_ind = int(res.shape[0] * (1 - threshold))
            test_start = res.iloc[test_start_ind][self.timestamp_column]
            res["is_test"] = res[self.timestamp_column] >= test_start
        else:
            res["is_test"] = res[self.timestamp_column] >= threshold

        if self.session_id_column:
            res = self._recalculate_with_session_id_column(res)

        train = res[~res["is_test"]].drop(columns=["is_test"])
        test = res[res["is_test"]].drop(columns=["is_test"])

        return train, test

    def _partial_split_spark(
        self, interactions: SparkDataFrame, threshold: Union[datetime, str, int]
    ) -> tuple[SparkDataFrame, SparkDataFrame]:
        if isinstance(threshold, float):
            dates = interactions.select(self.timestamp_column).withColumn(
                "_row_number_by_ts", sf.row_number().over(Window.orderBy(self.timestamp_column))
            )
            test_start = int(dates.count() * (1 - threshold)) + 1
            test_start = (
                dates.filter(sf.col("_row_number_by_ts") == test_start).select(self.timestamp_column).first()[0]
            )
            res = interactions.withColumn("is_test", sf.col(self.timestamp_column) >= test_start)
        else:
            res = interactions.withColumn("is_test", sf.col(self.timestamp_column) >= threshold)

        if self.session_id_column:
            res = self._recalculate_with_session_id_column(res)
        train = res.filter("is_test == 0").drop("is_test")
        test = res.filter("is_test").drop("is_test")

        return train, test

    def _partial_split_polars(
        self, interactions: PolarsDataFrame, threshold: Union[datetime, str, int]
    ) -> tuple[PolarsDataFrame, PolarsDataFrame]:
        if isinstance(threshold, float):
            test_start = int(len(interactions) * (1 - threshold)) + 1

            res = interactions.sort(self.timestamp_column).with_columns(
                (pl.col(self.timestamp_column).cum_count() >= test_start).alias("is_test")
            )
        else:
            res = interactions.with_columns((pl.col(self.timestamp_column) >= threshold).alias("is_test"))

        if self.session_id_column:
            res = self._recalculate_with_session_id_column(res)
        train = res.filter(~pl.col("is_test")).drop("is_test")
        test = res.filter("is_test").drop("is_test")

        return train, test

    def _core_split(self, interactions: DataFrameLike) -> list[DataFrameLike]:
        return self._partial_split(interactions, self.time_threshold)



"""
Most metrics require dataframe with recommendations
and dataframe with ground truth values —
which objects each user interacted with.
All input dataframes must have the same type.

- recommendations (Union[pandas.DataFrame, spark.DataFrame, Dict]):
    If the recommendations is instance of dict then key represents user_id, value represents tuple(item_id, score).
    If the recommendations is instance of Spark or Pandas dataframe
    then the names of the corresponding columns should be passed through the constructor of metric.
- ground_truth (Union[pandas.DataFrame, spark.DataFrame]):
    If ground_truth is instance of dict then key represents user_id, value represents item_id.
    If the recommendations is instance of Spark or Pandas dataframe
    then the names of the corresponding columns must match the recommendations.

Metric is calculated for all users, presented in ``ground_truth``
for accurate metric calculation in case when the recommender system generated
recommendation not for all users.  It is assumed, that all users,
we want to calculate metric for, have positive interactions.

Every metric is calculated using top ``K`` items for each user.
It is also possible to calculate metrics
using multiple values for ``K`` simultaneously.

Make sure your recommendations do not contain user-item duplicates
as duplicates could lead to the wrong calculation results.

- k (Union[Iterable[int], int]):
    a single number or a list, specifying the
    truncation length for recommendation list for each user

By default, metrics are averaged by users - ``replay.metrics.Mean``
but you can alternatively use ``replay.metrics.Median``.
You can get the median value of the confidence interval -
``replay.metrics.ConfidenceInterval`` for a given ``alpha``.
To calculate the metric value for each user separately in most metrics there is a parameter ``per user``.

To write your own aggregation kernel,
you need to inherit from the ``replay.metrics.CalculationDescriptor`` and redefine two methods (``spark``, ``cpu``).

For each metric, a formula for its calculation is given, because this is
important for the correct comparison of algorithms, as mentioned in our
`article <https://arxiv.org/abs/2206.12858>`_.
"""

from .base_metric import Metric
from .categorical_diversity import CategoricalDiversity
from .coverage import Coverage
from .descriptors import CalculationDescriptor, ConfidenceInterval, Mean, Median, PerUser
from .experiment import Experiment
from .hitrate import HitRate
from .map import MAP
from .mrr import MRR
from .ndcg import NDCG
from .novelty import Novelty
from .offline_metrics import OfflineMetrics
from .precision import Precision
from .recall import Recall
from .rocauc import RocAuc
from .surprisal import Surprisal
from .unexpectedness import Unexpectedness



from scipy.sparse import csr_matrix

from replay.models.extensions.ann.entities.nmslib_hnsw_param import NmslibHnswParam
from replay.models.extensions.ann.index_stores.base_index_store import IndexStore
from replay.models.extensions.ann.utils import create_nmslib_index_instance
from replay.utils import PandasDataFrame


class NmslibIndexBuilderMixin:
    """Provides nmslib index building method for different nmslib index builders"""

    @staticmethod
    def build_and_save_index(
        pdf: PandasDataFrame,
        index_params: NmslibHnswParam,
        index_store: IndexStore,
    ):
        """
        Builds nmslib index and saves it to index storage.
        This function is implemented to be able to test
        the `build_index_udf` internal functionality outside of spark,
        because pytest does not see this function call if the function is called by spark.
        :param pdf: pandas dataframe containing item similarities,
         with the following columns: `item_idx_one`, `item_idx_two`, `similarity`.
        :param index_params: index parameters as instance of NmslibHnswParam.
        :param index_store: index store
        :return:
        """
        creation_index_params = {
            "M": index_params.m,
            "efConstruction": index_params.ef_c,
            "post": index_params.post,
        }

        index = create_nmslib_index_instance(index_params)

        data = pdf["similarity"].values
        row_ind = pdf["item_idx_two"].values
        col_ind = pdf["item_idx_one"].values

        sim_matrix = csr_matrix(
            (data, (row_ind, col_ind)),
            shape=(
                index_params.items_count,
                index_params.items_count,
            ),
        )
        index.addDataPointBatch(data=sim_matrix)
        index.createIndex(creation_index_params)

        index_store.save_to_store(lambda path: index.saveIndex(path, save_data=True))



import logging
from typing import Any, Optional

from replay.utils import SparkDataFrame
from replay.utils.spark_utils import cache_temp_view, drop_temp_view


class RecommenderCommons:
    """
    Common methods and attributes of RePlay models for caching, setting parameters and logging
    """

    _logger: Optional[logging.Logger] = None
    cached_dfs: Optional[set] = None
    query_column: str
    item_column: str
    rating_column: str
    timestamp_column: str

    def set_params(self, **params: dict[str, Any]) -> None:
        """
        Set model parameters

        :param params: dictionary param name - param value
        :return:
        """
        for param, value in params.items():
            setattr(self, param, value)
        self._clear_cache()

    def _clear_cache(self):
        """
        Clear spark cache
        """

    def __str__(self):
        return type(self).__name__

    @property
    def logger(self) -> logging.Logger:
        """
        :returns: get library logger
        """
        if self._logger is None:
            self._logger = logging.getLogger("replay")
        return self._logger

    def _cache_model_temp_view(self, df: SparkDataFrame, df_name: str) -> None:
        """
        Create Spark SQL temporary view for df, cache it and add temp view name to self.cached_dfs.
        Temp view name is : "id_<python object id>_model_<RePlay model name>_<df_name>"
        """
        full_name = f"id_{id(self)}_model_{self!s}_{df_name}"
        cache_temp_view(df, full_name)

        if self.cached_dfs is None:
            self.cached_dfs = set()
        self.cached_dfs.add(full_name)

    def _clear_model_temp_view(self, df_name: str) -> None:
        """
        Uncache and drop Spark SQL temporary view and remove from self.cached_dfs
        Temp view to replace will be constructed as
        "id_<python object id>_model_<RePlay model name>_<df_name>"
        """
        full_name = f"id_{id(self)}_model_{self!s}_{df_name}"
        drop_temp_view(full_name)
        if self.cached_dfs is not None:
            self.cached_dfs.discard(full_name)



import pandas as pd
from scipy.sparse import csr_matrix


def get_csr_matrix(
    user_idx: pd.Series,
    vector_items: pd.Series,
    vector_ratings: pd.Series,
) -> csr_matrix:
    """Creates and returns csr matrix of user-item interactions"""
    return csr_matrix(
        (
            vector_ratings.explode().values.astype(float),
            (
                user_idx.repeat(vector_items.apply(lambda x: len(x))).values,
                vector_items.explode().values.astype(int),
            ),
        ),
        shape=(
            user_idx.max() + 1,
            vector_items.apply(lambda x: max(x)).max() + 1,
        ),
    )



from dataclasses import dataclass, field

from typing_extensions import Literal

from .base_hnsw_param import BaseHnswParam


@dataclass
class HnswlibParam(BaseHnswParam):
    """
    Parameters for hnswlib methods.

    For example,

    >>> HnswlibParam(space="ip",\
                     m=100,\
                     ef_c=200,\
                     post=0,\
                     ef_s=2000,\
        )
    HnswlibParam(space='ip', m=100, ef_c=200, post=0, ef_s=2000, dim=None, max_elements=None)

    or

    >>> HnswlibParam(space="ip",\
                     m=100,\
                     ef_c=200,\
                     post=0,\
                     ef_s=2000,\
        )
    HnswlibParam(space='ip', m=100, ef_c=200, post=0, ef_s=2000, dim=None, max_elements=None)

    The "space" parameter described on the page https://github.com/nmslib/hnswlib/blob/master/README.md#supported-distances
    Parameters "m", "ef_s" and "ef_c" are described at https://github.com/nmslib/hnswlib/blob/master/ALGO_PARAMS.md

    The reasonable range of values for `m` parameter is 5-100,
    for `ef_c` and `ef_s` is 100-2000.
    Increasing these values improves the prediction quality
    but increases index_time and inference_time too.

     We recommend using these settings:

    - m=16, ef_c=200 and ef_s=200 for simple datasets like MovieLens.
    - m=50, ef_c=1000 and ef_s=1000 for average quality with an average prediction time.
    - m=75, ef_c=2000 and ef_s=2000 for the highest quality with a long prediction time.

    note: choosing these parameters depends on the dataset
    and quality/time tradeoff.

    note: while reducing parameter values the highest range metrics
    like Metric@1000 suffer first.

    note: even in a case with a long training time,
    profit from ann could be obtained while inference will be used multiple times.
    """

    space: Literal["l2", "ip", "cosine"] = "ip"
    # Dimension of vectors in index
    dim: int = field(default=None, init=False)
    # Max number of elements that will be stored in the index
    max_elements: int = field(default=None, init=False)



"""
Contains classes for encoding categorical data

``LabelEncoderTransformWarning`` new category of warning for DatasetLabelEncoder.
``DatasetLabelEncoder`` to encode categorical features in `Dataset` objects.
"""

import warnings
from collections.abc import Iterable, Iterator, Sequence
from typing import Optional, Union

from replay.data import Dataset, FeatureHint, FeatureSchema, FeatureSource, FeatureType
from replay.preprocessing import LabelEncoder, LabelEncodingRule, SequenceEncodingRule
from replay.preprocessing.label_encoder import HandleUnknownStrategies


class LabelEncoderTransformWarning(Warning):
    """Label encoder transform warning."""


class DatasetLabelEncoder:
    """
    Categorical features encoder for the Dataset class
    """

    def __init__(
        self,
        handle_unknown_rule: HandleUnknownStrategies = "error",
        default_value_rule: Optional[Union[int, str]] = None,
    ) -> None:
        """
        :param handle_unknown_rule:
            When set to ``error`` an error will be raised in case an unknown label is present during transform.
            When set to ``use_default_value``, the encoded value of unknown label will be set
            to the value given for the parameter default_value.
            When set to ``drop``, the unknown labels will be dropped.
            Default: ``error``.
        :param default_value: Default value that will fill the unknown labels after transform.
            When the parameter handle_unknown is set to ``use_default_value``,
            this parameter is required and will set the encoded value of unknown labels.
            It has to be distinct from the values used to encode any of the labels in fit.
            If ``None``, then keep null.
            If ``int`` value, then fill by that value.
            If ``str`` value, should be \"last\" only, then fill by ``n_classes`` value.
            Default: ``None``.
        """
        self._handle_unknown_rule = handle_unknown_rule
        self._default_value_rule = default_value_rule
        self._encoding_rules: dict[str, LabelEncodingRule] = {}

        self._features_columns: dict[Union[FeatureHint, FeatureSource], Sequence[str]] = {}

    def fit(self, dataset: Dataset) -> "DatasetLabelEncoder":
        """
        Fits an encoder by the input Dataset for categorical features.

        :param dataset: the Dataset object.
        :returns: fitted DatasetLabelEncoder.
        :raises:
            AssertionError: if any of `dataset` categorical features contains
                invalid ``FeatureSource`` type.
        """

        self._fill_features_columns(dataset.feature_schema)
        for column, feature_info in dataset.feature_schema.categorical_features.items():
            encoding_rule_class = (
                SequenceEncodingRule if feature_info.feature_type == FeatureType.CATEGORICAL_LIST else LabelEncodingRule
            )
            encoding_rule = encoding_rule_class(
                column, handle_unknown=self._handle_unknown_rule, default_value=self._default_value_rule
            )
            if feature_info.feature_hint == FeatureHint.QUERY_ID:
                if dataset.query_features is None:
                    encoding_rule.fit(dataset.interactions)
                else:
                    encoding_rule.fit(dataset.query_features)
            elif feature_info.feature_hint == FeatureHint.ITEM_ID:
                if dataset.item_features is None:
                    encoding_rule.fit(dataset.interactions)
                else:
                    encoding_rule.fit(dataset.item_features)
            elif feature_info.feature_source == FeatureSource.INTERACTIONS:
                encoding_rule.fit(dataset.interactions)
            elif feature_info.feature_source == FeatureSource.QUERY_FEATURES:
                encoding_rule.fit(dataset.query_features)
            elif feature_info.feature_source == FeatureSource.ITEM_FEATURES:
                encoding_rule.fit(dataset.item_features)
            else:
                assert False, "Unknown feature source"  # pragma: no cover

            self._encoding_rules[column] = encoding_rule

        return self

    def transform(
        self,
        dataset: Dataset,
    ) -> Dataset:
        """
        Transforms the input Dataset categorical features by rules.

        :param dataset: The Dataset object.
        :returns: transformed dataset.
        """
        self._check_if_initialized()

        interactions = dataset.interactions
        query_features = dataset.query_features
        item_features = dataset.item_features

        for column, feature_info in dataset.feature_schema.categorical_features.items():
            if column not in self._encoding_rules:
                warnings.warn(
                    f"Cannot transform feature '{column}' as it was not present at the fit stage",
                    LabelEncoderTransformWarning,
                )
                continue

            encoding_rule = self._encoding_rules[column]

            if feature_info.feature_hint == FeatureHint.QUERY_ID:
                interactions = encoding_rule.transform(interactions)
                if query_features is not None:
                    query_features = encoding_rule.transform(query_features)
            elif feature_info.feature_hint == FeatureHint.ITEM_ID:
                interactions = encoding_rule.transform(interactions)
                if item_features is not None:
                    item_features = encoding_rule.transform(item_features)
            elif feature_info.feature_source == FeatureSource.INTERACTIONS:
                interactions = encoding_rule.transform(interactions)
            elif feature_info.feature_source == FeatureSource.QUERY_FEATURES:
                query_features = encoding_rule.transform(query_features)
            else:
                item_features = encoding_rule.transform(item_features)

        dataset_copy = Dataset(
            feature_schema=dataset.feature_schema,
            interactions=interactions,
            query_features=query_features,
            item_features=item_features,
            check_consistency=False,
            categorical_encoded=True,
        )

        return dataset_copy

    def fit_transform(self, dataset: Dataset) -> Dataset:
        """
        Fits an encoder and transforms the input Dataset categorical features.

        :param dataset: the Dataset object.
        :returns: transformed dataset.
        """
        return self.fit(dataset).transform(dataset)

    def get_encoder(self, columns: Union[str, Iterable[str]]) -> Optional[LabelEncoder]:
        """
        Get the encoder of fitted Dataset for columns.

        :param columns: columns to filter by.
        :returns: LabelEncoder.
        """
        self._check_if_initialized()

        columns_set: set[str]
        columns_set = {columns} if isinstance(columns, str) else {*columns}

        def get_encoding_rules() -> Iterator[LabelEncodingRule]:
            for column, rule in self._encoding_rules.items():
                if column in columns_set:
                    yield rule

        rules = list(get_encoding_rules())
        if len(rules) == 0:
            return None

        return LabelEncoder(rules)

    @property
    def query_id_encoder(self) -> LabelEncoder:
        """
        :returns: query id LabelEncoder.
        """
        query_id_column = self._features_columns[FeatureHint.QUERY_ID]
        encoder = self.get_encoder(query_id_column)
        assert encoder is not None
        return encoder

    @property
    def item_id_encoder(self) -> LabelEncoder:
        """
        :returns: item id LabelEncoder.
        """
        item_id_column = self._features_columns[FeatureHint.ITEM_ID]
        encoder = self.get_encoder(item_id_column)
        assert encoder is not None
        return encoder

    @property
    def query_and_item_id_encoder(self) -> LabelEncoder:
        """
        :returns: query id and item id LabelEncoder.
        """
        query_id_column = self._features_columns[FeatureHint.QUERY_ID]
        item_id_column = self._features_columns[FeatureHint.ITEM_ID]
        encoder = self.get_encoder(query_id_column + item_id_column)
        assert encoder is not None
        return encoder

    @property
    def interactions_encoder(self) -> Optional[LabelEncoder]:
        """
        :returns: interactions LabelEncoder.
        """

        interactions_columns = self._features_columns[FeatureSource.INTERACTIONS]
        return self.get_encoder(interactions_columns)

    @property
    def query_features_encoder(self) -> Optional[LabelEncoder]:
        """
        :returns: query features LabelEncoder.
        """
        query_features_columns = self._features_columns[FeatureSource.QUERY_FEATURES]
        return self.get_encoder(query_features_columns)

    @property
    def item_features_encoder(self) -> Optional[LabelEncoder]:
        """
        :returns: item features LabelEncoder.
        """
        item_features_columns = self._features_columns[FeatureSource.ITEM_FEATURES]
        return self.get_encoder(item_features_columns)

    def _check_if_initialized(self) -> None:
        if not self._encoding_rules:
            msg = "Encoder is not initialized"
            raise ValueError(msg)

    def _fill_features_columns(self, feature_info: FeatureSchema) -> None:
        self._features_columns = {
            FeatureHint.QUERY_ID: [feature_info.query_id_column],
            FeatureHint.ITEM_ID: [feature_info.item_id_column],
            FeatureSource.INTERACTIONS: feature_info.interaction_features.columns,
            FeatureSource.QUERY_FEATURES: feature_info.query_features.columns,
            FeatureSource.ITEM_FEATURES: feature_info.item_features.columns,
        }



from abc import abstractmethod
from typing import Union

import numpy as np
from scipy.stats import norm, sem

from replay.utils import PYSPARK_AVAILABLE, PolarsDataFrame, SparkDataFrame

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf


class CalculationDescriptor:
    """
    Abstract class for metric aggregation using given distribution by users
    """

    @property
    def __name__(self) -> str:
        return str(self.__class__.__name__)

    @abstractmethod
    def spark(self, distribution: SparkDataFrame):
        """
        Calculation on PySpark
        """

    @abstractmethod
    def cpu(self, distribution: Union[np.array, PolarsDataFrame]):
        """
        Calculation on cpu
        """


class Mean(CalculationDescriptor):
    """
    Calculation of the average value for a given distribution by users
    """

    def spark(self, distribution: SparkDataFrame):
        column_name = distribution.columns[0]
        return distribution.select(sf.avg(column_name)).first()[0]

    def cpu(self, distribution: Union[np.array, PolarsDataFrame]):
        if isinstance(distribution, PolarsDataFrame):
            return distribution.select(distribution.columns[0]).mean().rows()[0][0]
        return np.mean(distribution)


class PerUser(CalculationDescriptor):
    """
    Returns the distribution as is, for each user
    """

    def spark(self, distribution: SparkDataFrame):
        return distribution

    def cpu(self, distribution: Union[np.array, PolarsDataFrame]):
        return distribution


class Median(CalculationDescriptor):
    """
    Calculation of the medium value for a given distribution by users
    """

    def spark(self, distribution: SparkDataFrame):
        column_name = distribution.columns[0]
        return distribution.select(sf.expr(f"percentile_approx({column_name}, 0.5)")).first()[0]

    def cpu(self, distribution: Union[np.array, PolarsDataFrame]):
        if isinstance(distribution, PolarsDataFrame):
            return distribution.select(distribution.columns[0]).median().rows()[0][0]
        return np.median(distribution)


class ConfidenceInterval(CalculationDescriptor):
    """
    Calculating the average value of the confidence interval for a given distribution by users.
    To obtain the left boundary of the interval,
    it is necessary to subtract the result of this class from the average value.
    For the right border - add the average with the result of this class.
    """

    def __init__(self, alpha: float):
        self.alpha = alpha

    def spark(self, distribution: SparkDataFrame):
        column_name = distribution.columns[0]
        quantile = norm.ppf((1 + self.alpha) / 2)
        value = (
            distribution.agg(
                sf.stddev(column_name).alias("std"),
                sf.count(column_name).alias("count"),
            )
            .select(
                sf.when(
                    sf.isnan(sf.col("std")) | sf.col("std").isNull(),
                    sf.lit(0.0),
                )
                .otherwise(sf.col("std"))
                .cast("float")
                .alias("std"),
                "count",
            )
            .first()
        )
        return quantile * value["std"] / (value["count"] ** 0.5)

    def cpu(self, distribution: Union[np.array, PolarsDataFrame]):
        if isinstance(distribution, PolarsDataFrame):
            return self._polars(distribution)
        quantile = norm.ppf((1 + self.alpha) / 2)
        return quantile * sem(distribution)

    def _polars(self, distribution: PolarsDataFrame):
        column_name = distribution.columns[0]
        quantile = norm.ppf((1 + self.alpha) / 2)
        count = distribution.select(column_name).count().rows()[0][0]
        std = distribution.select(column_name).std().fill_null(0.0).fill_nan(0.0).rows()[0][0]
        return quantile * std / (count**0.5)



from replay.data.dataset import Dataset
from replay.utils import PYSPARK_AVAILABLE

from .base_rec import NonPersonalizedRecommender

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf


class PopRec(NonPersonalizedRecommender):
    """
    Recommend objects using their popularity.

    Popularity of an item is a probability that random user rated this item.

    .. math::
        Popularity(i) = \\dfrac{N_i}{N}

    :math:`N_i` - number of users who rated item :math:`i`

    :math:`N` - total number of users

    >>> import pandas as pd
    >>> from replay.data.dataset import Dataset, FeatureSchema, FeatureInfo, FeatureHint, FeatureType
    >>> from replay.utils.spark_utils import convert2spark
    >>> data_frame = pd.DataFrame(
    ...    {"user_id": [1, 1, 2, 2, 3, 4],
    ...     "item_id": [1, 2, 2, 3, 3, 3],
    ...     "rating": [0.5, 1, 0.1, 0.8, 0.7, 1]}
    ... )
    >>> data_frame
        user_id   item_id     rating
    0         1         1        0.5
    1         1         2        1.0
    2         2         2        0.1
    3         2         3        0.8
    4         3         3        0.7
    5         4         3        1.0

    >>> feature_schema = FeatureSchema(
    ...     [
    ...         FeatureInfo(
    ...             column="user_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.QUERY_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="item_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.ITEM_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="rating",
    ...             feature_type=FeatureType.NUMERICAL,
    ...             feature_hint=FeatureHint.RATING,
    ...         ),
    ...     ]
    ... )
    >>> interactions = convert2spark(data_frame)
    >>> dataset = Dataset(feature_schema, interactions)
    >>> res = PopRec().fit_predict(dataset, 1)
    >>> res.toPandas().sort_values("user_id", ignore_index=True)
        user_id   item_id     rating
    0         1         3       0.75
    1         2         1       0.25
    2         3         2       0.50
    3         4         2       0.50

    >>> res = PopRec().fit_predict(dataset, 1, filter_seen_items=False)
    >>> res.toPandas().sort_values("user_id", ignore_index=True)
        user_id   item_id     rating
    0         1         3       0.75
    1         2         3       0.75
    2         3         3       0.75
    3         4         3       0.75

    >>> res = PopRec(use_rating=True).fit_predict(dataset, 1)
    >>> res.toPandas().sort_values("user_id", ignore_index=True)
        user_id   item_id     rating
    0         1         3      0.625
    1         2         1      0.125
    2         3         2      0.275
    3         4         2      0.275

    """

    sample: bool = False

    def __init__(
        self,
        use_rating: bool = False,
        add_cold_items: bool = True,
        cold_weight: float = 0.5,
    ):
        """
        :param use_rating: flag to use rating values as is or to treat them as 1
        :param add_cold_items: flag to consider cold items in recommendations building
            if present in `items` parameter of `predict` method
            or `pairs` parameter of `predict_pairs` methods.
            If true, cold items are assigned rating equals to the less relevant item rating
            multiplied by cold_weight and may appear among top-K recommendations.
            Otherwise cold items are filtered out.
            Could be changed after model training by setting the `add_cold_items` attribute.
        : param cold_weight: if `add_cold_items` is True,
            cold items are added with reduced rating.
            The rating for cold items is equal to the rating
            of a least relevant item multiplied by a `cold_weight` value.
            `Cold_weight` value should be in interval (0, 1].
        """
        self.use_rating = use_rating
        super().__init__(add_cold_items=add_cold_items, cold_weight=cold_weight)

    @property
    def _init_args(self):
        return {
            "use_rating": self.use_rating,
            "add_cold_items": self.add_cold_items,
            "cold_weight": self.cold_weight,
        }

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        agg_func = sf.countDistinct(self.query_column).alias(self.rating_column)
        if self.use_rating:
            agg_func = sf.sum(self.rating_column).alias(self.rating_column)

        self.item_popularity = (
            dataset.interactions.groupBy(self.item_column)
            .agg(agg_func)
            .withColumn(self.rating_column, sf.col(self.rating_column) / sf.lit(self.queries_count))
        )

        self.item_popularity.cache().count()
        self.fill = self._calc_fill(self.item_popularity, self.cold_weight, self.rating_column)



import importlib
import logging
import sys
from abc import abstractmethod
from collections.abc import Iterable
from typing import Any, Optional, Union

from replay.data import Dataset
from replay.models.common import RecommenderCommons
from replay.utils import ANN_AVAILABLE, PYSPARK_AVAILABLE, FeatureUnavailableError, SparkDataFrame

from .index_builders.base_index_builder import IndexBuilder

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf

    from replay.utils.spark_utils import get_top_k_recs, return_recs

    from .index_stores.spark_files_index_store import SparkFilesIndexStore

logger = logging.getLogger("replay")


class ANNMixin(RecommenderCommons):
    """
    This class overrides the `_fit_wrap` and `_predict_wrap` methods of the base class,
    adding an index construction in the `_fit_wrap` step
    and an index inference in the `_predict_wrap` step.
    """

    index_builder: Optional["IndexBuilder"] = None

    def init_index_builder(self, index_builder: Optional[IndexBuilder] = None) -> None:
        if index_builder is not None and not ANN_AVAILABLE:
            err = FeatureUnavailableError(
                "`index_builder` can only be provided when all ANN dependencies are installed."
            )
            if sys.version_info >= (3, 11):  # pragma: py-lt-311
                err.add_note(
                    "To enable ANN, ensure you have both 'hnswlib' and 'fixed-install-nmslib' packages installed."
                )
            raise err
        elif isinstance(index_builder, IndexBuilder):
            self.index_builder = index_builder
        elif isinstance(index_builder, dict):
            self.init_builder_from_dict(index_builder)

    @property
    def _use_ann(self) -> bool:
        """
        Property that determines whether the ANN (index) is used.
        If `True`, then the index will be built (at the `fit` stage)
        and index will be inferred (at the `predict` stage).
        """
        return self.index_builder is not None

    @abstractmethod
    def _configure_index_builder(self, interactions: SparkDataFrame) -> tuple[SparkDataFrame, dict]:
        """Implementation of this method must return dictionary
        with arguments for for an index builder's`build_index` method.

        Args:
            interactions: DataFrame with interactions

        Returns:
            vectors: DataFrame[item_idx int, vector array<double>] or DataFrame[vector array<double>].
        Column names in dataframe can be anything.
            ann_params: Dictionary with arguments to build index. For example: {
            "id_col": "item_idx",
            "features_col": "item_factors",
            ...
        }

        """

    def _fit_wrap(
        self,
        dataset: Dataset,
    ) -> None:
        """Wrapper extends `_fit_wrap`, adds construction of ANN index by flag.

        Args:
            dataset: historical interactions with query/item features
                ``[user_id, item_id, timestamp, rating]``
        """
        super()._fit_wrap(dataset)

        if self._use_ann:
            vectors, ann_params = self._configure_index_builder(dataset.interactions)
            self.index_builder.build_index(vectors, **ann_params)

    @abstractmethod
    def _get_vectors_to_infer_ann_inner(self, interactions: SparkDataFrame, queries: SparkDataFrame) -> SparkDataFrame:
        """Implementations of this method must return a dataframe with user vectors.
        User vectors from this method are used to infer the index.

        Args:
            interactions: DataFrame with interactions
            queries: DataFrame with queries

        Returns: DataFrame[user_idx int, vector array<double>] or DataFrame[vector array<double>].
        Vector column name in dataframe can be anything.
        """

    def _get_vectors_to_infer_ann(
        self, interactions: SparkDataFrame, queries: SparkDataFrame, filter_seen_items: bool
    ) -> SparkDataFrame:
        """This method wraps `_get_vectors_to_infer_ann_inner`
        and adds seen items to dataframe with user vectors by flag.

        Args:
            interactions: DataFrame with interactions
            queries: DataFrame with queries
            filter_seen_items: flag to remove seen items from recommendations based on ``interactions``.

        Returns:

        """
        queries = self._get_vectors_to_infer_ann_inner(interactions, queries)

        # here we add `seen_item_idxs` to filter the viewed items in UDFs (see infer_index_udf)
        if filter_seen_items:
            user_to_max_items = interactions.groupBy(self.query_column).agg(
                sf.count(self.item_column).alias("num_items"),
                sf.collect_set(self.item_column).alias("seen_item_idxs"),
            )
            queries = queries.join(user_to_max_items, on=self.query_column)

        return queries

    @abstractmethod
    def _get_ann_infer_params(self) -> dict[str, Any]:
        """Implementation of this method must return dictionary
        with arguments for `_infer_ann_index` method.

        Returns: dictionary with arguments to infer index. For example: {
            "features_col": "user_vector",
            ...
        }

        """

    def _predict_wrap(
        self,
        dataset: Optional[Dataset],
        k: int,
        queries: Optional[Union[SparkDataFrame, Iterable]] = None,
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
        filter_seen_items: bool = True,
        recs_file_path: Optional[str] = None,
    ) -> Optional[SparkDataFrame]:
        dataset, queries, items = self._filter_interactions_queries_items_dataframes(dataset, k, queries, items)

        if self._use_ann:
            vectors = self._get_vectors_to_infer_ann(dataset.interactions, queries, filter_seen_items)
            ann_params = self._get_ann_infer_params()
            inferer = self.index_builder.produce_inferer(filter_seen_items)
            recs = inferer.infer(vectors, ann_params["features_col"], k)
        else:
            recs = self._predict(
                dataset,
                k,
                queries,
                items,
                filter_seen_items,
            )

        if not self._use_ann:
            if filter_seen_items and dataset.interactions:
                recs = self._filter_seen(recs=recs, interactions=dataset.interactions, queries=queries, k=k)

            recs = get_top_k_recs(recs, k=k, query_column=self.query_column, rating_column=self.rating_column).select(
                self.query_column, self.item_column, self.rating_column
            )

        output = return_recs(recs, recs_file_path)
        self._clear_model_temp_view("filter_seen_queries_interactions")
        self._clear_model_temp_view("filter_seen_num_seen")
        return output

    def _save_index(self, path):
        self.index_builder.index_store.dump_index(path)

    def _load_index(self, path: str):
        self.index_builder.index_store = SparkFilesIndexStore()
        self.index_builder.index_store.load_from_path(path)

    def init_builder_from_dict(self, init_meta: dict):
        """Inits an index builder instance from a dict with init meta."""

        # index param entity instance initialization
        module = importlib.import_module(init_meta["index_param"]["module"])
        class_ = getattr(module, init_meta["index_param"]["class"])
        index_params = class_(**init_meta["index_param"]["init_args"])

        # index builder instance initialization
        module = importlib.import_module(init_meta["builder"]["module"])
        class_ = getattr(module, init_meta["builder"]["class"])
        index_builder = class_(index_params=index_params, index_store=None)

        self.index_builder = index_builder
